#include "../include/parser.hpp"

CParser::CParser() { 
    m_variables.emplace("ans", CRational(new CBigInt(0)));
    m_continue = 1;
    m_saver.reset(new CTextsaver(m_variables));
    // CRational::termPrecision = 4;
}

CParser::~CParser() { clear(); }

uint64_t CParser::getOperatorPriority( TOKEN t ) const {
    uint64_t ret = 0;
    switch (t)
    {
    case DEREFERENCE:
        ret = 8;
        break;
    case LN:
        ret = 7;
        break;
    case EXP:
        ret = 6;
        break;
    case DIVIDE:
        ret = 5;
        break;
    case MULTIPLY:
        ret = 4;
        break;
    case MOD:
        ret = 3;
        break;
    case ADD:
        ret = 2;
        break;
    case SUBTRACT:
        ret = 1;
        break;
    default:
        // throw "unresolvable syntax error";
        break;
    }
    return ret;
}

CParser::CMD CParser::recognizeCommand( uint64_t commandPos ) const {
                                  // position in raw lexems
    CParser::CMD ret = UNKNOWN_CMD;
    char c = m_input[m_rawLexems[commandPos].second];
    // std::cout << c;
    if (!std::islower(c))
        return ret;
    
    switch (c) {
    case 'q':
        ret = QUIT_CMD;
        break;
    case 'h':
        ret = HELP_CMD;
        break;
    case 'p':
        ret = PRECISSION_CMD;
        break;
    case 's':
        ret = SAVE_CMD;
        break;
    case 'l':
        ret = LOAD_CMD;
        break;
    case 'r':
        ret = ROUND_CMD;
        break;
    default:
        return ret;
        break;
    }
    
    char nextChar = m_input[m_rawLexems[commandPos].second + 1];
    if (m_input.size() == (uint64_t)m_rawLexems[commandPos].second
     || nextChar == ' '
     || nextChar == '\0') // short command
        return ret;

    std::string comStr;
    switch (ret) {
    case QUIT_CMD:
        comStr = m_input.substr( m_rawLexems[commandPos].second, sizeof("quit") - 1);
        ret = comStr == "quit" ? QUIT_CMD : UNKNOWN_CMD;
        break;
    case HELP_CMD:
        comStr = m_input.substr( m_rawLexems[commandPos].second, sizeof("help") - 1);
        ret = comStr == "help" ? HELP_CMD : UNKNOWN_CMD;
        break;
    case PRECISSION_CMD:
        comStr = m_input.substr( m_rawLexems[commandPos].second, sizeof("precision") - 1);
        ret = comStr == "precision" ? PRECISSION_CMD : UNKNOWN_CMD;
        break;
    case SAVE_CMD:
        comStr = m_input.substr( m_rawLexems[commandPos].second, sizeof("save") - 1);
        ret = comStr == "save" ? SAVE_CMD : UNKNOWN_CMD;
        break;
    case LOAD_CMD:
        comStr = m_input.substr( m_rawLexems[commandPos].second, sizeof("load") - 1);
        ret = comStr == "load" ? LOAD_CMD : UNKNOWN_CMD;
        break;
    case ROUND_CMD:
        comStr = m_input.substr( m_rawLexems[commandPos].second, sizeof("round") - 1);
        ret = comStr == "round" ? ROUND_CMD : UNKNOWN_CMD;
        break;
    default:
        break;
    }
    return ret;
}

CParser::CMD CParser::parseCommand( uint64_t &startPos ) const {
                                  // position in raw lexems
    if (m_rawLexems[startPos].first == COMMAND) {

        return recognizeCommand(startPos + 1);
    } else {
        startPos = 1;
        return NONE_CMD;
    }
}

bool CParser::executeCommand( CParser::CMD cmd ) {
    switch (cmd) {
    case QUIT_CMD:
        m_continue = 0;
        m_output = "quitting..";
        return false;
        break;
    case HELP_CMD:
        helpCommand();
        return false;
        break;
    case PRECISSION_CMD:
        if (m_processedLexems.size() != 7)
            return false;
        try {
            CRational::termPrecision = std::stoull(m_input.substr(m_rawLexems[4].second));
            m_output = "precision set";
            if (CRational::termPrecision == 0) {
                m_output += ", zero will propably break exp and ln calculations!";
            }
        } catch ( std::exception &e ) {
            m_output = "invalid precision, recommended values are 1-10";
        }
        return false;
        break;
    case SAVE_CMD:
        if (m_processedLexems.size() != 7)
            return false;
        if (m_saver->save( m_input.substr(m_rawLexems[4].second) ))
            m_output = "save successful";
        else
            m_output = "save failed";
        return false;
        break;
    case LOAD_CMD:
        if (m_processedLexems.size() != 7)
            return false;
        if (m_saver->load( m_input.substr(m_rawLexems[4].second) ))
            m_output = "load successful";
        else
            m_output = "load failed";
        return false;
        break;
    case ROUND_CMD:
        if (m_processedLexems.size() != 8 || 
           (m_processedLexems[4].first != NUMBER && m_processedLexems[5].first != NUMBER))
            return false;
        try {
            uint64_t decimals = std::stoull( m_input.substr(m_rawLexems[4].second) );
            //todo check
            m_output = m_processedLexems[5].second->round(decimals);
        } catch (std::exception &e) {
            m_output = "rounding failed";
        }
        return false;
        break;
    case UNKNOWN_CMD:
        m_output = "unknown command. try !help";
        return false;
        break;
    case NONE_CMD:
        break;
    }
    return true;
}

bool CParser::helpCommand() {
    m_output =  "!h !help - for this quick help\n";
    m_output += "!q !quit - to quit\n";
    m_output += "!p !precision x - to set number of terms for aproximating logarith and exponencial (positive int or LOW=4/HIGH=8, values below 1 or above 10 are not recommended)\n";
    m_output += "!s !save - to save variables (this is not automatic on quit!)\n";
    m_output += "!l !load - to load them back (this is not automatic on quit!)\n\n";
    m_output += "!r !round x y - to round y to x decimals\n";

    m_output += "start by typing numbers and operators\n";
    m_output += "exact operation supported for racional numbers:\n";
    m_output += "addition (+), subtraction (-), multiplication (*), division (/)\n";
    m_output += "example usage:\n";
    m_output += "3.25 * ( -4/3 + 3/4 )\n\n";

    m_output += "aproximation of logarith (&) and exponencial (^)\n";
    m_output += "example usage:\n";
    m_output += "(3)&(27) represents logartih with base 3 out of 27\n\n";

    m_output += "(everything is case sensitive!)\n";
    m_output += "run \"dot ./doc/grammar.dot -T x11\" to see grammar guide\n";

    return true;
}

void CParser::clear() {
    m_rawLexems.clear();
    for( auto &i : m_processedLexems )
        if (i.second)
            delete i.second;
    m_processedLexems.clear();
    m_output = "SYNTAX ERROR try !help";
}

CParser::TOKEN CParser::recognize( char c ) const {
    if (c == ' ')
        return CParser::TOKEN::WHITESPACE;
    if (c >= '0' && c <= '9')
        return CParser::TOKEN::NUMBER;
    if ((c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z'))
        return CParser::TOKEN::LABEL;

    CParser::TOKEN ret;
    switch (c) {
    case '!':
        ret = COMMAND;
        break;
    case '$':
        ret = DEREFERENCE;
        break;
    case '(':
        ret = OPEN;
        break;
    case ')':
        ret = CLOSE;
        break;
    case '&':
        ret = LN;
        break;
    case '^':
        ret = EXP;
        break;
    case '/':
        ret = DIVIDE;
        break;
    case '*':
        ret = MULTIPLY;
        break;
    case '%':
        ret = MOD;
        break;
    case '+':
        ret = ADD;
        break;
    case '-':
        ret = SUBTRACT;
        break;
    case '.':
        ret = DOT;
        break;
    case '=':
        ret = SAVE;
        break;
    default:
        ret = SYNTAX_ERROR;
        break;
    }
    return ret;
}

CRational *CParser::parseNumber( uint64_t i, bool isDecimal = false, bool hasSign = false ) const {
    CBigInt *num = new CBigInt(0);
    CBigInt *den;
    std::string rawNumber = readToken(i);
    if (isDecimal) {
        rawNumber += readToken(i+2);
        int64_t decimalLen = m_rawLexems[i+2].second;
        for (;decimalLen < m_rawLexems[i+3].second && m_input[decimalLen] != ' '; ++decimalLen);
        decimalLen -= m_rawLexems[i+2].second;

        if (decimalLen <= 10)
            den = new CBigInt(std::pow(10, decimalLen));
        else {
            CBigInt *tmp = new CBigInt( 10 );
            den = tmp->power(decimalLen);
            delete tmp;
        }
        
    } else {
        den = new CBigInt(1);
    }
    hasSign ^= 1;
    hasSign ^= 1;
    // convert ascii
    for ( auto &c : rawNumber) {
        c -= 48;
    }

    uint64_t insertPosition = 0;
    while (!std::all_of(rawNumber.begin(), rawNumber.end(), [](char c) { return c == 0; })) {
        num->setBit(insertPosition++, stringDivideDecimal(rawNumber));
    }
    
    return new CRational(num, den);
}

bool CParser::stringDivideDecimal( std::string &s ) { // returns next bit for binary conversion
    uint16_t additive = 0;
    uint16_t nextAdditive;
    for (char &c : s) {
        nextAdditive = (c % 2) * 5;
        c = c / 2 + additive;
        std::swap(additive, nextAdditive);
    }
    return additive;
}

bool CParser::tokenize() {
    m_rawLexems.emplace_back(START, -1);
    m_rawLexems.emplace_back(OPEN, -1);
    for (uint64_t i = 0; i < m_input.size(); ++i) {
        TOKEN type = recognize(m_input[i]);
        if (type == SYNTAX_ERROR)
            return false;
        if (type == WHITESPACE) {
            if (m_rawLexems.back().first != WHITESPACE)
                m_rawLexems.emplace_back(WHITESPACE, i);
            continue;
        }
        if (type == NUMBER) {
            if (m_rawLexems.back().first != NUMBER)
                    m_rawLexems.emplace_back(NUMBER, i);
            continue;
        }
        if (type == LABEL) {
            if (m_rawLexems.back().first != LABEL)
                m_rawLexems.emplace_back(LABEL, i);
            continue;
        }
        if (type == COMMAND || type == SAVE) {
            // if (m_rawLexems.size() == 2)
                m_rawLexems.emplace_back(type, i);
            // else 
            //     return false;
            continue;
        }
        m_rawLexems.emplace_back(type, i);
    }
    m_rawLexems.emplace_back(CLOSE, m_input.size());
    m_rawLexems.emplace_back(END, m_input.size());
    return m_rawLexems.size() > 4;
}

std::string CParser::readToken( uint64_t i ) const {
    std::string ret;
    for ( int64_t j = m_rawLexems[i].second; j < m_rawLexems[i+1].second && m_input[j] != ' '; ++j ) {
        ret.push_back(m_input[j]);
    }
    return ret;
}

bool CParser::grammarPreprocess() {
    // strip whitespaces
    m_rawLexems.erase(
        std::remove_if(
            m_rawLexems.begin(),
            m_rawLexems.end(),
            [](const std::pair<CParser::TOKEN, int64_t>& lexem) {
                return lexem.first == WHITESPACE;
            }
        ),
        m_rawLexems.end()
    );

    m_saveInto.clear();
    if (m_rawLexems[2].first == LABEL) {
        uint64_t varLen = m_rawLexems[2].second;
        for (; (m_input[varLen] >= 'a' && m_input[varLen] <= 'z') || (m_input[varLen] >= 'A' && m_input[varLen] <= 'Z'); ++ varLen);
        varLen -= m_rawLexems[2].second;
        m_saveInto = m_input.substr(m_rawLexems[2].second, varLen);

        if (m_rawLexems[3].first != SAVE) {
            m_output = "SYNTAX ERROR try " + m_saveInto + " = " +
            (m_input.substr(m_rawLexems[3].second).size() ? m_input.substr(m_rawLexems[3].second) : "(1 + 2) * 3");
            return false;
        }
        m_rawLexems.erase( m_rawLexems.begin() + 2, m_rawLexems.begin() + 4);
    }

    // process numbers
    uint64_t i = 0;
    for (; i < m_rawLexems.size(); ++i) {
        bool numberHasDecimals = false;
        bool numberHasDot = false;
        bool numberHasSign = false;
        bool numberNegative = false;
        if (m_rawLexems[i].first == WHITESPACE)
            continue;

        if (m_rawLexems[i].first == DOT) {
            if (m_rawLexems[i+1].first == NUMBER) {
                m_processedLexems.emplace_back(NUMBER, parseNumber(i+1, false, false));

                uint64_t decimalLen = m_rawLexems[i+1].second;
                for (; m_input[decimalLen] >= '0' &&  m_input[decimalLen] <= '9'; ++decimalLen);
                decimalLen -= m_rawLexems[i+1].second;

                if (decimalLen <= 10)
                    m_processedLexems.back().second->m_denominator.reset( new CBigInt(std::pow(10, decimalLen)));
                else {
                    CBigInt *tmp = new CBigInt( 10 );
                    m_processedLexems.back().second->m_denominator.reset( tmp->power(decimalLen) );
                    delete tmp;
                }
                ++i;
                continue;
            }
            return false;
        }

        if (m_rawLexems[i].first == NUMBER) {
            if (m_processedLexems.back().first == NUMBER)
                return false;
            numberHasDot = (m_rawLexems[i+1].first == DOT);

            numberHasDecimals = numberHasDot && m_rawLexems[i+2].first == NUMBER;
            numberHasSign =  ((m_rawLexems[i-1].first == ADD || m_rawLexems[i-1].first == SUBTRACT) 
                             &&
                                (m_processedLexems[m_processedLexems.size() -2].first != NUMBER
                             &&  m_processedLexems[m_processedLexems.size() -2].first != CLOSE));

            if (numberHasSign && !m_processedLexems.empty()) {
                numberNegative = m_processedLexems.back().first == SUBTRACT;

                m_processedLexems.pop_back();

            }
            
            m_processedLexems.emplace_back(NUMBER, parseNumber(i, numberHasDecimals, numberHasSign));
            m_processedLexems.back().second->m_numerator->m_negative = numberNegative;
            i += numberHasDot;
            i += numberHasDecimals;
            continue;
        }

        if (m_rawLexems[i].first == DEREFERENCE) {
            if (m_rawLexems[i+1].first != LABEL) {
                m_output = "invalid dereference";
                return false;
            }
            uint64_t varLen = m_rawLexems[i+1].second;
            for (; (m_input[varLen] >= 'a' && m_input[varLen] <= 'z') || (m_input[varLen] >= 'A' && m_input[varLen] <= 'Z'); ++ varLen);
            varLen -= m_rawLexems[i+1].second;
            std::string variableName = m_input.substr(m_rawLexems[i+1].second, varLen);
            auto it = m_variables.find(variableName);
            if (it == m_variables.end()) {
                m_output = "unknown variable \"" + variableName + "\"";
                return false;
            }
            CRational *value = new CRational((*it).second);
            m_processedLexems.emplace_back(NUMBER, value);
            ++i;
            continue;
        }
        m_processedLexems.emplace_back(m_rawLexems[i].first, nullptr);

    }
    return true;
}

std::string CParser::numberToString( CRational *r ) const {
    std::string ret = r->m_numerator->toString();
    CBigInt one(1);
    if (r->m_denominator->cmp(&one))
        ret += " / " + r->m_denominator->toString();
    return ret;
}

COperator *CParser::spawnOperator( CParser::TOKEN t ) const {
    COperator *ret = nullptr;
    switch (t) {
    case ADD:
        ret = new CAddition();
        break;
    case SUBTRACT:
        ret = new CSubtraction();
        break;
    case MULTIPLY:
        ret = new CMultiplication();
        break;
    case DIVIDE:
        ret = new CDivision();
        break;
    case MOD:
        ret = new CModulo();
        break;
    case EXP:
        ret = new CExponencial();
        break;
    case LN:
        ret = new CLogarith();
        break;
    default:
        break;
    }
    return ret;
}

CPrintable *CParser::parse( uint64_t &readIndex ) {
    enum STATE {
        FIRST_OPERAND,
        OPERATOR,
        SECOND_OPERAND
    };
    STATE parserState = FIRST_OPERAND;
    CPrintable *firstOperand   = nullptr;
    // CPrintable *secondOperand  = nullptr;
    COperator *currentOperator = nullptr;
    COperator *lastOperator    = nullptr;
    uint64_t lastPriority = 0;
    
    ++readIndex; // skip first bracket
    while (m_processedLexems[readIndex].first != CLOSE) {
        if (m_processedLexems[readIndex].first == OPEN) {
            if (parserState == FIRST_OPERAND) {
                firstOperand = parse(readIndex);
            } else if (parserState == SECOND_OPERAND) {
                currentOperator->m_secondArgument.reset( parse(readIndex) );
            } else {
                goto fail;
            }
            parserState = OPERATOR;
            ++readIndex;
            continue;
        }

        switch (parserState) {

        case FIRST_OPERAND:
            if (m_processedLexems[readIndex].first == OPEN) {
                firstOperand = parse(readIndex);
                parserState = OPERATOR;
                break;
            }
            if (m_processedLexems[readIndex].first == NUMBER) {
                firstOperand = m_processedLexems[readIndex].second;
                m_processedLexems[readIndex].second = nullptr;
                parserState = OPERATOR;
                break;
            }
            goto fail;
        
        case OPERATOR:
            if (m_processedLexems[readIndex].first == CLOSE)
                goto sucess;

            currentOperator = spawnOperator(m_processedLexems[readIndex].first);
            if (!currentOperator)
                goto fail;
            
            if (lastOperator) {
                uint64_t thisPriority = getOperatorPriority(m_processedLexems[readIndex].first);

                if (lastPriority <= thisPriority) {
                    currentOperator->m_firstArgument.reset( lastOperator->m_secondArgument.release() );
                    lastOperator->m_secondArgument.reset( currentOperator );
                } else {
                    currentOperator->m_firstArgument.reset( lastOperator );
                    lastOperator = currentOperator;
                }
            } else {
                lastPriority = getOperatorPriority(m_processedLexems[readIndex].first);
                lastOperator = currentOperator;
                lastOperator->m_firstArgument.reset( firstOperand );
            }
            parserState = SECOND_OPERAND;
            break;

        case SECOND_OPERAND:
            if (m_processedLexems[readIndex].first == OPEN) {
                currentOperator->m_secondArgument.reset( parse(readIndex) );
                parserState = OPERATOR;
                break;
            }
            if (m_processedLexems[readIndex].first == NUMBER) {
                currentOperator->m_secondArgument.reset( m_processedLexems[readIndex].second );
                m_processedLexems[readIndex].second = nullptr;
                parserState = OPERATOR;
                break;
            }
            goto fail;
            break;
        }
        ++readIndex;
    }

    sucess:
    if ((currentOperator && (!currentOperator->m_firstArgument.get() || !currentOperator->m_secondArgument.get())) ||
        (lastOperator && (!lastOperator->m_firstArgument.get() || !lastOperator->m_secondArgument.get())))
        throw std::logic_error("parsing failed");
    if (lastOperator)
        return lastOperator;
    if (firstOperand)
        return firstOperand;

    fail:
    throw std::logic_error("parsing failed");
}

void CParser::convert( const std::string &s ) {
    clear();
    m_input = s;
    if (!tokenize())
        return;
    if (!grammarPreprocess())
        return;
    uint64_t readPos = 2;
    CMD command = parseCommand(readPos);
    if (command != NONE_CMD && !m_saveInto.empty())
        return;
    if (!executeCommand( command ))
        return;

    try {
    m_head.reset( parse( readPos ) );
    } catch (const std::exception &e) {
        m_output = "Caught exception: " + std::string(e.what());
        return;
    } catch (...) {
        m_output = "Caught unknown exception";
        return;
    }
    
    if (m_processedLexems[++readPos].first != END) {
        m_output = "bad parenthesis parity";
        return;
    }
    if (!m_head.get()) {
        m_output = "parsing failed";
        return;
    }
    try {
        auto evaluated = m_head->evaluate();

        m_output = numberToString(evaluated);
        auto ans = m_variables.find("ans");
        (*ans).second.m_numerator.reset( evaluated->m_numerator.release() );
        (*ans).second.m_denominator.reset( evaluated->m_denominator.release() );

        if (!m_saveInto.empty()) {
            auto var = m_variables.find(m_saveInto);

            if (var == m_variables.end()) {
                var = m_variables.emplace(m_saveInto, CRational( new CBigInt(0) )).first;
            }
            (*var).second.m_numerator.reset( new CBigInt( *(*ans).second.m_numerator.get() ));
            (*var).second.m_denominator.reset( new CBigInt( *(*ans).second.m_denominator.get() ));
        }
        delete evaluated;
    } catch (const std::exception &e) {
        m_output = "Caught exception: " + std::string(e.what());
    } catch (...) {
        m_output = "Caught unknown exception";
    }
}