#pragma once
#include "rational.hpp"
#include "operator.hpp"

/**
 * @class CExponencial
 * @brief The CExponencial class represents an exponential operation.
 * It inherits from the COperator class.
 */
class CExponencial : public COperator {
public:
    /**
     * @brief Constructor for CExponencial class.
     * @param a The base of the exponential operation.
     * @param b The exponent of the exponential operation.
     */
    CExponencial( CPrintable *a, CPrintable *b );

    /**
     * @brief Default constructor for CExponencial class.
     */
    CExponencial() = default;

    /**
     * @brief Destructor for CExponencial class.
     */
    ~CExponencial() = default;

    /**
     * @brief Evaluates the exponential operation.
     * @return A CRational pointer representing the result of the exponential operation.
     */
    virtual CRational *evaluate() override;
private:
};
