#pragma once
#include "rational.hpp"
#include "operator.hpp"

/**
 * @brief The CModulo class represents a modulo operation.
 * It inherits from the COperator class.
 */
class CModulo : public COperator {
public:
    /**
     * @brief Constructor for CModulo class.
     * @param a The first argument of the modulo operation.
     * @param b The second argument of the modulo operation.
     */
    CModulo( CPrintable *a, CPrintable *b );

    /**
     * @brief Default constructor for CModulo class.
     */
    CModulo() = default;

    /**
     * @brief Destructor for CModulo class.
     */
    ~CModulo() = default;

    /**
     * @brief Evaluates the modulo operation.
     * @return A CRational pointer representing the result of the modulo operation.
     * @throws An exception if the modulo operation is applied to rationals.
     */
    virtual CRational *evaluate() override;
private:
};
