#pragma once
#include "printable.hpp"

/**
 * @brief The COperator class represents an abstract operator.
 * It inherits from the CPrintable class.
 */
class COperator : public CPrintable {
public:
    /**
     * @brief Constructor for COperator class.
     * @param a The first argument of the operator.
     * @param b The second argument of the operator.
     */
    COperator( CPrintable *a, CPrintable *b );

    /**
     * @brief Default constructor for COperator class.
     */
    COperator() = default;

    /**
     * @brief Default destructor for COperator class.
     */
    virtual ~COperator() = default;

    /**
     * @brief Evaluates the operands and then itself.
     * @return A CRational pointer representing the result of the operator.
     */
    virtual CRational *evaluate() = 0;

// protected:
    /**
     * @brief The first argument of the operator.
     */
    std::unique_ptr<CPrintable> m_firstArgument;

    /**
     * @brief The second argument of the operator.
     */
    std::unique_ptr<CPrintable> m_secondArgument;
};