#pragma once
#include "../include/saver.hpp"
#include "../include/parser.hpp"

/**
 * @class CTextsaver
 * @brief A class for saving and loading variables in text format.
 */
class CTextsaver : public CSaver {
private:
    CRational *parseNumber( const std::string &s ) const;
public:
    /**
     * @brief Constructor for CTextsaver.
     * @param variablesMap A reference to the map of variables to be saved.
     */
    CTextsaver(std::map<std::string, CRational> &variablesMap);

    /**
     * @brief Destructor for CTextsaver.
     */
    ~CTextsaver();

    /**
     * @brief Saves the variables to a file in text format.
     * @param fileName The name of the file to save the variables.
     * @return True if the variables were successfully saved, false otherwise.
     */
    virtual bool save( const std::string &fileName ) override;

    /**
     * @brief Loads variables from a file in text format.
     * @param fileName The name of the file to load the variables from.
     * @return True if the variables were successfully loaded, false otherwise.
     */
    virtual bool load( const std::string &fileName ) override;
};